// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: CalPNResponse
// Author: Hassan Siddiqui
//
// Description: 
// Provides a clean interface to CAL response data.
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: CalPNResponse.cc,v 1.17 2007/06/06 12:42:14 eojero Exp $

/* General */
#include <string>
#include <iostream>
#include <vector>

/* SAS-specific */
#include "Dal.h"
#include "CalServer.h"
#include "Redistribution.h"
#include "Tracer.h"

// Local classes
#include "CalPNResponse.h"

/*****************************
 *
 * class CalPNResponse
 *
 * Interface to CAL response data.
 *
 *****************************/


inline void CalPNResponse::init( PixCoords cenRaw )
{
  
  PROC;

  /* Currently set state for default EPN settings */
  // calServer.state()->node()->set(Xmm::PRIMARY_NODE);
  
  /* Get response information */
  CanonicalRedist * rmf = 0; rmf = calServer.getAtom(rmf);

  /* Set source centre in raw coordinates */
  rmf->at(cenRaw.x(), cenRaw.y());

  /* Set PN line number */
  errstr 
    << "CalPNResponse: response model appropriate for line " << cenRaw.y()
    << Sas::message(Msg::App, Msg::Verbose);

  rmf->at(cenRaw.y());   // extract redist spectrum appl. at CCD line

  _rmf = rmf;
  
  ENDPROC;
}

/* Constructors */
/* Provides the entire range of 
 * redistribution available from the CAL */
CalPNResponse::CalPNResponse ( PixCoords cenRaw,
			       unsigned lowerPI,
			       unsigned nOutputChannels,
			       unsigned outputChannelWidth,
                               EventGrades eventGrade)
{
  init(cenRaw);

  _eventGrade=eventGrade;

  createWithCalEbins(lowerPI, 
		     nOutputChannels, 
		     outputChannelWidth);
}  

/* Provides a restricted range */
CalPNResponse::CalPNResponse ( PixCoords cenRaw,
			       unsigned lowerPI,
			       unsigned nOutputChannels,
			       unsigned outputChannelWidth,
			       const BinnedAxis * ebins,
                               EventGrades eventGrade)
{
  init(cenRaw);

  _eventGrade=eventGrade;

   createWithUserEbins( lowerPI, 
		       nOutputChannels, 
		       outputChannelWidth,
		       ebins);
}
/*
   Method: spatialAverage
   Purpose: Uses a detector map to calculate the spatially averaged
            redistribution parameters
*/
void CalPNResponse::spatialAverage(DetectorMap *_detmap) const
{
  PROC;
                                                                                
   /* Find the weighted mean RAW-Y position somwhow */
    PixCoords pcoords = _detmap->wtmean_centre_raw();

   /* Reset the spatial centre of the redistribution function */
    _rmf->at(pcoords.y());
                                                                                
  ENDPROC
}

/* Destructor */
CalPNResponse::~CalPNResponse()
{
  PROC;
  delete _eboundsAxis; _eboundsAxis = 0;
  ENDPROC;
}

