// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// 
// Module: ConcreteRmfDataServer
// Author: Hassan Siddiqui
//
// Description: Data pool which is used during arf generation
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: ConcreteRmfDataServer.cc,v 1.18 2013/02/15 15:04:00 rsaxton Exp $

/* General */
#include <iostream>

/* SAS-specific */
#include "Tracer.h"
#include "Xmm.h"

/* Local */
#include "ConcreteRmfDataServer.h"
#include "CorePosData.h"
#include "DummyResponse.h"
#include "CalMOSResponse.h"
#include "CalPNResponse.h"
#include "NullResponseCorrector.h"
#include "PiledUpResponseCorrector.h"
#include "TablePhotonDistribution.h"
#include "NullPhotonDistribution.h"
#include "SeekFixedMatrixColumn.h"
#include "FixedMatrixColumn.h"
#include "LowMemFixedMatrixColumn.h"
#include "VarMatrixColumn.h"
#include "BinnedProxyResponse.h"
#include "ExceptionErrorHandler.h"

// =================================================================

// class ConcreteRmfDataServer

// Data pool which is used during rmf generation

// =================================================================

/* Constructors */ 
ConcreteRmfDataServer::ConcreteRmfDataServer( const string& instrument,
					      unsigned      tlmin,
					      unsigned      tlmax,
					      unsigned      nBinnedChannels,
					      unsigned      binfactor,
					      unsigned      offset,
					      double        energyLowkeV,
					      double        energyHighkeV,
					      unsigned      nEnergyBins,
					      bool          withebins, 
                                              CorePosData   *corePosData,
                                              const string& submode,    
                                              const string& format,
                                              bool          correctForPileUp,
                                              const string&       eventfile,
                                              PixelIterator *pxitr):
  _instrument(instrument),
  _tlmin(tlmin),
  _tlmax(tlmax),
  _fixedlength(true),
  _withebins(withebins)
{
      PROC;

       errstr << "Pattern code="<<corePosData->patternCode()<<" "<<
                 "Submode = "<<submode
        << Sas::message(Msg::App, Msg::Verbose);

      /* Convert energy ranges to eV */
      double energyLow = energyLowkeV * 1000.;
      double energyHigh = energyHighkeV * 1000.;

      /* Determine o/p dataset format */
      if (format=="var")
  	{
	  _fixedlength=false;
  	}
      
      /* Create PhotonDistribution object */
      _photonDistribution = 
	new NullPhotonDistribution();

      /* Create Frame object */
      _frame = new Frame(_photonDistribution);

      /* Get instrument ID */
      _instrumentID = Xmm::toInstrument(_instrument);

      if (_withebins)
	{
	  _ebins = new UniformilyBinnedAxis(energyLow, 
					    energyHigh,
					    nEnergyBins);
	  
	  if (Xmm::toInstrument(instrument) == Xmm::EPN)
	    {
	      
	      _response = new CalPNResponse( corePosData->centre_raw(),
					     offset,
					     nBinnedChannels,
					     binfactor,
					     _ebins,
                                             corePosData->patternCode());
	    }
	  else
	    {
	      _response = new CalMOSResponse( offset,
					      nBinnedChannels,
					      binfactor,
					      _ebins,
                                              corePosData->patternCode());
	    }
	}
      else
	{
	  
	  if (Xmm::toInstrument(instrument) == Xmm::EPN)
	    {
	      _response = new CalPNResponse( corePosData->centre_raw(),
					     offset,
					     nBinnedChannels,
					     binfactor,
                                             corePosData->patternCode());
	    }
	  else
	    {
	      _response = new CalMOSResponse( offset,
					      nBinnedChannels,
                                              binfactor,
                                              corePosData->patternCode());

	    }
	}

     /* Add a response corrector for pile-up effects if requested */
      if (correctForPileUp)
         _responseCorrector = new PiledUpResponseCorrector( _response,
                                   eventfile, corePosData, offset, 
                                   binfactor, pxitr);
      else
         _responseCorrector = new NullResponseCorrector( _response);


      ENDPROC;
    };


/* Destructors */
ConcreteRmfDataServer::~ConcreteRmfDataServer()
{
      PROC;
      
      delete _responseCorrector;  _responseCorrector  = 0;
      delete _response;           _response           = 0;
      //      delete _response_in;        _response           = 0;
      if (_withebins)
	{
	  delete _ebins;
	}
      delete _frame;              _frame              = 0;
      delete _photonDistribution; _photonDistribution = 0;
    
      ENDPROC;
};

/* Accessors */
  MatrixColumn *  ConcreteRmfDataServer::matrixColumn( Table * table,
						       const vector<unsigned long>& sizeVect ) const
{
  PROC;
  if (!_fixedlength)
    {
      ALTRETURNV( new VarMatrixColumn(table) );
    }

//    string memmod(getenv("SAS_MEMORY_MODEL"));
		
//    errstr << " Got memory model to be " << memmod
//  	 << Sas::message(Msg::App, Msg::Verbose);

//    if ( memmod=="low" )
//         {
//  	 ALTRETURNV(new LowMemFixedMatrixColumn(table, sizeVect));
//         }
  
  /* Memory model is high, or not defined */
  //  RETURN new FixedMatrixColumn(table, sizeVect);
  //  RETURN new LowMemFixedMatrixColumn(table, sizeVect);
  RETURN new SeekFixedMatrixColumn(table, sizeVect);
  
}



Xmm::Instrument ConcreteRmfDataServer::instrumentID()      const 
    {
      PROC; RETURN _instrumentID;
    }

const string    ConcreteRmfDataServer::instrument()        const 
    { 
      PROC; RETURN _instrument;
    }

const PhotonDistribution *  ConcreteRmfDataServer::photonDistibution() const 
{ 
  PROC; RETURN _photonDistribution;
}

const GenericResponse    *  ConcreteRmfDataServer::response()          const 
{ 
  PROC; RETURN _response;
}

const ResponseCorrector  *  ConcreteRmfDataServer::responseCorrector() const 
    { 
      PROC; RETURN _responseCorrector;
    }

  const Frame              * ConcreteRmfDataServer::frame()             const  
    { 
      PROC; RETURN _frame;
    }

  /* Spectrum-related properties */
  unsigned                         ConcreteRmfDataServer::tlmin()             const
    {
      PROC; RETURN _tlmin;
    }

  unsigned                         ConcreteRmfDataServer::tlmax()             const
    {
      PROC; RETURN _tlmax;
    }
