// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: DetectorMap
// Author: Richard Saxton
//
// Description:
// A detector map
//
// Comments:
//
// Modifications: See ChangeLog
//
/* General classes */
#include <string>
#include <iostream>
#include <vector>
#include <cmath>

/* THIS CLASS CALLS attcalc, SO RUN AS METATASK */
#include <metatask.h>
#include "attcalc.h"

/* SAS-specific classes */
#include "DetectorMap.h"
#include "DataSubspace.h"
#include "TablePixelIterator.h"
#include "CompactTablePixelIterator.h"
#include "ArrayPixelIterator.h"
#include "DatafileAttributes.h"
#include "PsfModel.h"
#include "PsfmodelPixelIterator.h"
#include "DummyPixelIterator.h"
#include "DataFilter.h"
#include "CheckExpr.h"

/********************************
 *
 * class DetectorMap
 *
 * A map to be used for spatial averaging
 *
 ********************************/

/* Constructor */

DetectorMap::DetectorMap(PixelIterator * const filteredDetmap, const PositionUtils *putils):
_filteredDetmap(filteredDetmap),
_rawSet(false)
{
  PROC

    errstr<<"Constructing DetectorMap"
            << Sas::message(Msg::App, Msg::Verbose);

  /* Extract the individual pixel positions and values into arrays */
    FOREACH(_filteredDetmap)
    {
      PixelElement pixel = _filteredDetmap->normalisedItem();

      _detx.push_back(pixel.xCentre());
      _dety.push_back(pixel.yCentre());
      _weight.push_back(pixel.value());

    }

   /* Find the raw pixel values */
    setRaw(putils);

  ENDPROC;
}

/* Destructor */
DetectorMap::~DetectorMap()
{ };

/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => trailing const removed
 */
PixelIterator *  DetectorMap::filteredMap()
{ PROC;

   RETURN _filteredDetmap;
}

/* Method: detX()
   Purpose: returns a vector of x positions
*/
vector<double> DetectorMap::detX() const
{ PROC;

 RETURN _detx;
}

/* Method: detY()
   Purpose: returns a vector of y positions
*/
vector<double> DetectorMap::detY() const
{ PROC;

 RETURN _dety;
}

/* Method: weight()
   Purpose: returns a vector of weights normalised to 1.0
*/
vector<double> DetectorMap::weight() const
{ PROC;

 RETURN _weight;
}

/* Method: rawX()
   Purpose: returns a vector of x chip positions
*/
vector<unsigned short > DetectorMap::rawX() const
{ PROC;

 RETURN _rawx;
}

/* Method: rawy()
   Purpose: returns a vector of y chip positions
*/
vector<unsigned short> DetectorMap::rawY() const
{ PROC;

 RETURN _rawy;
}

/* Method: setRaw()
   Purpose: sets up arrays of rawx, rawy positions
*/
void DetectorMap::setRaw(const PositionUtils *putils)
{
  PROC;

   /* Calculate the raw pixel coords for each detector map pixel
      if not already done */
    if (! _rawSet) {

       unsigned short xpix=0, ypix=0;

   /* Loop over each record in the filtered map and find the RAW pixel vals */
       GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

       for (unsigned int i=0; i<_detx.size(); i++)
       {
        /* Convert from detector to camera coords */
          CamCoords testpixel = putils->detxy005ToCamCoord2(
                                                    _detx[i], _dety[i]);
        /* Convert from camera to RAW coords. If this pixel doesn't lie on a CCD
           do something */
          if ( ct->onCcd(testpixel.x(), testpixel.y()) == Xmm::CCD_NONE )
          {
           /* Source is off-chip so set the raw coordinates as 32.0, the Y
              value calculated from the DET-Y position */

            xpix = 32;
            double rawY = 199.0 - abs(_dety[i] + 1000) * 199.0 / 16000.0;
            ypix = UNT(rawY);

           // Artificially make rawy lie within the chip
            ypix=max(INT(ypix),1);
            ypix=min(INT(ypix),199);

          } else
          {
           /* Source is on chip so convert coords */
            PixCoords testpixCoord =
                        putils->camCoord2ToPixCoord(testpixel);
            xpix = testpixCoord.x();
            ypix = testpixCoord.y();
          }

          _rawx.push_back(xpix);
          _rawy.push_back(ypix);

       } // end of loop over pixels in filtered map

       _rawSet = true;
    }

    ENDPROC;
}


/* Return the mean source position in raw coordinates */
PixCoords DetectorMap::wtmean_centre_raw() const
{

  if (! _rawSet) {
     errstr << "weighted mean centre requested before raw coordinates "
            << "have been set up - programming error"
            << Sas::error("NoRawCoords");
  }

  PixCoords pixCoord(0,0);

  /* Loop over each pixel in the detector map */
  double xtot=0.0, ytot=0.0, totweight=0.0, dxpix=0.0;
  unsigned xpix=0, ypix=0;

  for (unsigned int i=0; i<_rawx.size(); i++)
  {

    /* Add up the weighted average */
     xtot += _rawx[i] * _weight[i];
     ytot += _rawy[i] * _weight[i];
     totweight += _weight[i];

     errstr<<"Mean centre: "<<_rawx[i]<<" "<< _rawy[i]<<" "<<_weight[i]
           << Sas::message(Msg::App, Msg::Noisy);
   }

  /* Average the pixels  and find the closest integer pixel*/
   dxpix = (xtot / totweight);
   xpix = UNT(xtot / totweight +0.5);
   ypix = UNT(ytot / totweight +0.5);

   pixCoord = PixCoords(xpix,ypix);

   errstr<<"Mean centre raw: "<<xpix<<" "<<ypix<<" "<<dxpix
         <<" "<<UNT(dxpix)
         << Sas::message(Msg::App, Msg::Verbose);

  RETURN  pixCoord;
}



