// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Module: DummyResponse
// Author: Hassan Siddiqui
//
// Description:
// Provides a clean interface to Dummy response data.
//
// Comments:
//
// Modifications: See ChangeLog
//
// $Id: DummyResponse.cc,v 1.18 2009/09/03 14:18:07 aibarra Exp $

/* General classes */
#include <string>
#include <iostream>
#include <vector>
#include <cmath>

/* SAS-specific classes */
#include "Dal.h"
#include "Tracer.h"
#include "NumericalRange.h"

/* Local classes */
#include "DummyResponse.h"

/********************************
 *
 * class DummyResponse
 *
 * Interface to Dummy response data.
 *
 ********************************/


/* Constructor */

DummyResponse::DummyResponse( unsigned int nchannels,
			      double energyLow,
			      double energyHigh,
			      unsigned int numberOfEbins
			      ):
  _energyLow(energyLow),
  _energyHigh(energyHigh),
  _numberOfEbins(numberOfEbins)
{
  PROC;
  _firstChannel = 0;
  _lastChannel  = nchannels-1;

  // Generate ebins
  //_ebins.erase(_ebins.begin(), _ebins.end());
  _ebins = 0;

  double deltaEbin = (_energyHigh - _energyLow )/ double(numberOfEbins);

  _ebins = new UniformilyBinnedAxis(0., numberOfEbins, deltaEbin);

  double deltaEbound = (_energyHigh - _energyLow )/ double(numberOfChannels());

  _ebounds = new UniformilyBinnedAxis(0., numberOfChannels(), deltaEbound);
  ENDPROC;
}

/* Destructor */
DummyResponse::~DummyResponse()
{ delete _ebins; _ebins = 0; delete _ebounds; _ebounds = 0; };


/** Channel range accessors */
/* AI gcc4 porting
 * error: type qualifiers ignored on function return type
 * => leading const removed
 */
unsigned int  DummyResponse::firstChannel()     const { PROC; RETURN _firstChannel; }
unsigned int  DummyResponse::lastChannel()      const { PROC; RETURN _lastChannel; }
/* AI gcc4 porting
 * end
 */

/** Response Accessors */
unsigned int DummyResponse::numberOfEbins()    const { PROC; RETURN _numberOfEbins; }

const Bin DummyResponse::ebins (unsigned int row) const
{ PROC; RETURN _ebins->bin(row); }

const vector<double>  DummyResponse::responseVector ( unsigned int energy_iterator ) const
{
  PROC;

  errstr << "DummyResponse::responseVector: getting row "
	 <<  energy_iterator
	 << " from data of " << _numberOfEbins
	 << " rows "
	 << Sas::message(Msg::App, Msg::Verbose);

  //      assert ( ebounds().count(energy_iterator) > 0 );

  Bin tmp_ebin = _ebins->bin( energy_iterator );

  RETURN responseVector( 0.5 * (tmp_ebin.lo() + tmp_ebin.hi()));
}

const vector<double> DummyResponse::responseVector ( double energy ) const
{
      PROC;

      // Generate a gaussian function,
      // with peak determined by input energy
      unsigned int meanChannel = energyToPha( energy );

      errstr << " mean channel for energy "
	     <<  energy
	     << " is "
	     << meanChannel
	     << Sas::message(Msg::App, Msg::Verbose);

      vector<double> gaussianResponse;

      /* Compute energy resolution (FWHM) of gaussian
       * Use dE = k * E^{0.5}
       * For MOS, k = 1.7 (approx); PN similar.
       */

      double sigma = _ENERGYRES_CONST * sqrt(energy);

      errstr << " mean channel for energy "
	     <<  energy
	     << " is "
	     << meanChannel
	     << " sigma is "
	     << sigma
	     << Sas::message(Msg::App, Msg::Verbose);

      // assert (sigma > FLT64_MIN);

      for (unsigned int i = 0; i < numberOfChannels(); ++i)
	{
	  // Compute exponent term, taking care of unsigned int conversions
	  double exp_term = (double(i) - double(meanChannel)) / sigma;
	  double value = exp( (-0.5) * exp_term * exp_term );

	  gaussianResponse.push_back( value );
	}

      // NOTE: Should normalise response here!!

      RETURN gaussianResponse;
    }

  /** Ebounds Accessors */

const BinnedAxis * DummyResponse::ebounds()  const
    { PROC; RETURN _ebounds; }


unsigned int  DummyResponse::energyToPha( double energy ) const
    {
      // assert ( energy >= _energyLow && energy <= _energyHigh );


      // Get upper and lower bounds of countsspectrum

      RETURN (unsigned int) (_firstChannel +
			     ( (energy - _energyLow)/
			       (_energyHigh - _energyLow) *
			       (_lastChannel - _firstChannel )) );

    }




