// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: PiledUpResponseCorrector
// Author: Hassan Siddiqui
//
// Description: 
// Corrects response for pile-up
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: PiledUpResponseCorrector.cc,v 1.6 2014/04/21 12:59:28 aibarra Exp $

// General classes

#include <iostream>
#include <vector>
#include <map>

// SAS-specific classes

#include "Dal.h"
#include "Tracer.h"
#include "Xmm.h"
#include "Random.h"

// Local Classes
#include "PiledUpResponseCorrector.h"
#include "GenericResponse.h"
#include "Frame.h"
#include "ResponseCorrector.h"
#include "CorePosData.h"
// #include "RmfUtils.h"
#include "PattFracs.h"
#include "PsfImage.h"
//gcc4.8.2 portability change
#include <unistd.h>

// -----------------------------------------------------------------------------------

// class PiledUpResponseCorrector

// Corrects response row for pile-up

// -----------------------------------------------------------------------------------

//class PiledUpResponseCorrector : public ResponseCorrector
//{

  /* Constructor */

  PiledUpResponseCorrector::PiledUpResponseCorrector( 
                            const GenericResponse  * response,
                            const string & eventfile,
                            CorePosData   *corePosData,
                            unsigned      chanoffset,  unsigned chanwidth,
                            PixelIterator * pxitr) :
        _response(response),
        _chanOffset(chanoffset), _chanWidth(chanwidth), 
        _pxitr(pxitr)
    {
      PROC;

     /* Perform initialisation here */
      _cenRaw = corePosData->centre_raw();
      _cenTel = corePosData->centre_telcoord();
      _cenDet = corePosData->centre_detcoord();
      _eventGrade = corePosData->patternCode();
      _spectrum_times = corePosData->getTimeList();

     // Read charges from the raw data file (this also fills in the
     // vector _framesToFill which is used later when constructing
     // the test events.
     // MAYBE SUBSET IF TOO MANY ??
      ObsEvents obsevents_all = readEvents(eventfile);

      errstr<<"Read Events X[5]:"<<obsevents_all.x[5]
            << Sas::message(Msg::App, Msg::Noisy);

     // Remove charges which lie outside the extraction region
      ObsEvents obsevents_region = extractCharges(obsevents_all);
 
     // DBG
      //for (int n=0; n<20; ++n)  
        //  cout<<"N: "<<n<<" "<<obsevents_region.x[n]<<obsevents_region.frame[n];
     // Normalise the frames to keep the count rate constant
      _obsevents = normalise_frames(obsevents_region);

   // Find frame ranges from the indices
    // _framesToFill = getFramesToFill(frame, index_range, totalFrames);

      _framesToFill = getFramesToFill(_obsevents, _N1EventFrames);
      //_framesToFill = getFramesToFill(_obsevents, _totalFrames);

      errstr<<"Final event used: "<<_obsevents.frame[_obsevents.frame.size()-1]
            <<" "<<_obsevents.charge[_obsevents.frame.size()-1]
            <<" tot used: "<<_obsevents.frame.size()
            <<Sas::message(Msg::App, Msg::Verbose);

      ENDPROC;
     }

  
  PiledUpResponseCorrector::~PiledUpResponseCorrector()
    {
      PROC;
      ENDPROC;
    }

  /* Calculate the pile-up corrected response for this input energy */ 
  const vector<double> PiledUpResponseCorrector::correctedResponse( 
                               const Interval<double>& ebin0,
                               const vector<double>& response0) const
    {
      PROC;

     // Define the central bin energy
      double energy0 = 0.5 * ( ebin0.lo() + ebin0.hi() );

     /*
      * Create a structure of test events to add to the observation charges.
      * One event per exposed frame, written into a structure with :
      *
      * - position randomised over the pixels
      * contained in the filtered extraction region (weighted by the PSF
      * at this energy.
      *
      * - charge randomised from the RMF at this energy
      *
      * - pattern randomised from the pattern fractions at this energy
     */
      ObsEvents *testevents = testEvents(energy0, response0);

     // Merge the observation and test events
      ObsEvents mergedevents = mergeEvents(_obsevents, testevents);

     // Use the output test events to calculate a new response
      vector<double> outResponse = makeRmfRow(mergedevents, 
                                           testevents->charge.size());

     // Clean up this iteration
      delete testevents; testevents=0;

      return outResponse;
   }


/** Method: testEvents()
* Purpose: Sets a structure of events, 1 per input frame, at a given
           input energy, randomised over the PSF 
*/
  PiledUpResponseCorrector::ObsEvents * PiledUpResponseCorrector::testEvents( 
                 double energy0, const vector<double> response0) const
    {

      ObsEvents *testevents = new ObsEvents;

     // Set randomisation seed
      randGen.reset();
      randGen.seed(getpid());

     // Set a local value to hold the number of events wanted
      int nevents = _framesToFill.size();

      errstr<<"Processing "<<nevents<<" test events with energy "<<energy0
            << Sas::message(Msg::App, Msg::Verbose);

     // Find the pattern fractions at this position and energy
      PattFracs *patternFractions = new PattFracs(_cenRaw); //PERFORMANCE

      vector<double> fracs;
      if (_eventGrade == CombinedEventGrades::SINGLE ||
            _eventGrade == CombinedEventGrades::SD)
         fracs.push_back(patternFractions->fraction(CombinedEventGrades::SINGLE,
                                                energy0));
      if (_eventGrade == CombinedEventGrades::DOUBLE ||
            _eventGrade == CombinedEventGrades::SD)
         fracs.push_back(patternFractions->fraction(CombinedEventGrades::DOUBLE,
                                                energy0));

     // Calculate a random pattern for each event
      vector<int> values(2); values[0]=INT(CombinedEventGrades::SINGLE);
      values[1]=INT(CombinedEventGrades::DOUBLE);
      testevents->pattern = findrand(nevents, fracs, values);

     // Randomise the total charge using the unpiled-up response fn
      //testevents->charge = findrand(nevents, response0);
     ///////////// TRY THE SLOWER METHOD /////////////////////
     //////// NOT CLEAR WHETHER DISTRIBUTING or RANDOMISING is BEST /////
      testevents->charge = findrand2(nevents, response0.data(), 
                                                 response0.size());

      errstr<<"Random charges "<<testevents->charge[0]
            << Sas::message(Msg::App, Msg::Noisy);

     /* Randomise the X,Y position on the chip from the PSF and the
        extraction region */
      vector<CoordPair> xypos = getChipPosition(nevents, 
                                        energy0, patternFractions);

     // Set PI channel, frame number and event id for each test event
      for (int n=0; n<nevents; ++n) 
      { 
         // Convert charge from RMF channel into PI channels
          testevents->charge[n] = testevents->charge[n]*_chanWidth 
                                             + _chanOffset;
         // Set frame number
          testevents->frame.push_back(_framesToFill[n]);

         // Set event ID to be simply the event number
          testevents->event_id.push_back(n);

         // Set X and Y positions
          testevents->x.push_back(xypos[n].x());
          testevents->y.push_back(xypos[n].y());
          testevents->weight.push_back(1);
      }

      errstr<<"TestEvents[0]:"<<testevents->charge[0]<<" "
            <<testevents->frame[0]<<" "<<testevents->x[0]
            << Sas::message(Msg::App, Msg::Noisy);

     // Tidy up
      delete patternFractions; patternFractions=0;
 
      return testevents;
    }

/*
* Method: getChipPosition()
* Purpose: Spread the events over the extraction region weighting
*          by the PSF
*/
  const vector<CoordPair> PiledUpResponseCorrector::getChipPosition(
                                int nevents, double energy0, 
                                PattFracs *patternFractions) const
  {

    vector<CoordPair> xypos;

  /* Get the PSF for this energy */
    PsfImage *psfim = new PsfImage(energy0, patternFractions, 
                               _cenTel.theta(), _cenTel.phi());

  /* Find the pixels in the PSF which are used in the extraction region.
     Set others to zero and renormalise */
    psfim->cleanRegion(_pxitr, _cenDet);

  // Normalise remaining pixels to 1.0
    psfim->normalise();

  // Read the PSF image (for pattern 0)
    //Image<float> pim = psfim->image(0);
    Image<float> &pim = psfim->image(0);

  // Convert 2-d image into 1-d array
    int nx = pim.xExtent(), ny = pim.yExtent();

    //int midx = INT(nx/2.0), midy = INT(ny/2.0);

    //cout<<"PSF centre: "<<pim[midx-1][midy-1]<<" "<<pim[midx-1][midy]<<" "
      //  <<pim[midx-1][midy+1]<<" "<<pim[midx][midy-1]<<" "
        //<<pim[midx][midy]<<" "<<pim[midx][midy+1]<<" "<<endl;

   // Find a random pixel for each event
    vector<int> position = findrand2(nevents, pim.data(), pim.size());

    errstr<<"POS: "<<position[0]<<" "<<position[1]<<" "<<position[2]
            << Sas::message(Msg::App, Msg::Noisy);

   // Get the plate scale for this instrument
    MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
    double plateX = mds->get("PLATE_SCALE_X");
    double plateY = mds->get("PLATE_SCALE_Y");
    calServer.releaseAtom(mds);


    int ix=0, iy=0;
    for (int n=0; n<nevents; ++n) {

       // Calculate the X and Y pixel position in PSF Units (arcsecs)
        ix = floor(position[n] / ny); // Modulus function
        iy = position[n] - ix * ny;

       // Convert this into raw chip coordinates
        ix = floor ( ( ix - floor(nx/2) ) / plateX + 0.5) + _cenRaw.x();
        iy = floor ( ( iy - floor(ny/2) ) / plateY + 0.5) + _cenRaw.y();
      
       // DEBUG
        if (n<10) 
           errstr<<"IX: "<<ix<<" "<<iy<<" "<<plateX<<" "<<round((nx)/2)
                 <<" "<<_cenRaw.x()<<" "<<_cenRaw.y()
                 << Sas::message(Msg::App, Msg::Noisy);

        xypos.push_back(CoordPair(ix,iy));
    }
   
    delete psfim; psfim=0;

    return xypos;
  }

/*
* Method: findrand()
* Purpose: Find random element from probability array
*/
  const vector<int> PiledUpResponseCorrector::findrand(int N, 
                      vector<double> probs, vector<int> values) const
  {
     // Get random element positions
      vector<int> pos = findrand(N, probs);

     // Loop over each required output
      vector<int> result;
      for (int i=0; i<N; ++i) {
          result.push_back(values[pos[i]]);
      }          
          
      return result;
   }

/*
* Method: findrand(int N, vector<double> probs)
* Purpose: Assign values to an array based on a probability array
*/
  const vector<int> PiledUpResponseCorrector::findrand(int N, 
                                             vector<double> probs) const
  {
      vector<int> result;

     // Set the size of the probabilities array
      int Np = probs.size();

     // Find sum of probabilities and the cumulative probability 
      double sumProb=0.0;
      for (int i=0; i<Np; ++i)
          sumProb += probs[i];

     // Use probabilities to divide N events

     // Loop over each probability
      //int filled=0, nev=0;
      int nev=0;
      for (int i=0; i<Np; ++i)
       {
          //filled=filled+nev;

       // Calculate how many events belong to this probability bin
          nev = floor(N * probs[i] / double(sumProb) +0.5);

       // Fill them
          //for (int n=filled; n<(filled+nev); n++)
          for (int n=0; n<nev; n++)
              result.push_back(i);
       }
      // If less than we need then randomise n more
       int filled = result.size();

       if (filled < N) 
          for (int n=filled; n<N; n++)
              result.push_back(result[filled-1]);

     // Now randomise positions of vector 
       std::random_shuffle ( result.begin(), result.end() );

     // Remove if too many
      if (filled > N)
         for (int n=filled-1; n>=N; n--) {
             result.erase (result.begin()+n);
         }

      return result;
   }

/*
* Method: findrand2(int N, vector<double> probs)
* Purpose: Find random element from probability array
*/
  //const vector<int> PiledUpResponseCorrector::findrand2(int N, 
    //                                         vector<float> probs) const
 template<class T> 
  const vector<int> PiledUpResponseCorrector::findrand2(int N, 
                                             T *probs, unsigned Np) const
  {
      vector<int> result(N);

     // Set the size of the probabilities array
      //unsigned Np = probs.size();

     // Find sum of probabilities and the cumulative probability
      float sumProb=0.0;
      for (unsigned i=0; i<Np; ++i)
          sumProb += probs[i];

      vector<float> cumProb(Np);
      cumProb[0]=probs[0];
      for (unsigned i=1; i<Np; ++i)
          cumProb[i]=cumProb[i-1]+probs[i];

     // Generate a random number between zero and the sum of the probs
      randGen.min(0.0);
      randGen.max(sumProb);

     // Loop over each needed value
      for (int i=0; i<N; ++i) {

         // Generate a random value
          double p = randGen.rand();

         // Handle the extreme cases
          if (p>=cumProb[Np-2])
             result[i] = Np-1;
          else if (p<=cumProb[0])
             result[i] = 0;
          else
          { // Binary chop to a solution
             int a = floor(Np/2);
             int da = max(INT(floor(Np/4)),1);

             while (p < cumProb[a] || p > cumProb[a+1])
              {
                if (p < cumProb[a])
                   a = a - da;
                else
                   a = a + da;

                da = max(INT(floor(da/2)),1);
              }

              result[i] = a+1;
          }
      //    cout<<"P: "<<p<<" "<<result[i]<<" "<<cumProb[result[i]]<<" "<<cumProb[result[i]+1]<<endl;
      }

      return result;
   }

/*
* Method: normalise_frames(ObsEvents obsevents)
* Purpose: This removes frames such that the total number of events per 
*          frame is the same as in the input event file.
*
*          If the input is N(1), N(2), N(3), N(4) frames where N is the 
*          number of frames with that number of events then the output 
*          should be:
*   
*          N(1) = N(2) ; N(2) = N(3); N(3) = N(4) ; N(4) = 0
*
*          If N is a descreasing function then events will be removed.
*          If it increases then events will be duplicated
*
*/
  PiledUpResponseCorrector::ObsEvents 
             PiledUpResponseCorrector::normalise_frames(
                        PiledUpResponseCorrector::ObsEvents obsevents) 
    {

     // Define output structure
     ObsEvents frame_normalised_events;

     unsigned N = obsevents.frame.size(); // N = number of input charges

     _sum_of_frame_weights = 0; // Zero the sum of the frame weights global var.

     // Define a vector to hold the number of charges in each frame
     vector<int> ncharges_in_frame(obsevents.frame[N-1]+1,0);

     for (unsigned i=0; i<N; ++i)
         ++ ncharges_in_frame[obsevents.frame[i]];

     // Define a vector to hold the number of apparent events in each frame
     vector<double> pattot_in_frame(obsevents.frame[N-1]+1,0.0);
     vector<double> good_pattot_in_frame(obsevents.frame[N-1]+1,0.0);
     // and the number of "good" events in each frame
     vector<int> nevents_in_frame(obsevents.frame[N-1]+1,0);
     vector<int> good_nevents_in_frame(obsevents.frame[N-1]+1,0);

     // Count the events from the pattern values. This is where
     // ambiguity and error is introduced into the method, as we don't know
     // whether two adjacent charges are due to one double-pixel event or
     // two single-pixel events.
     for (unsigned i=0; i<N; ++i) {
        if (abs(obsevents.pattern[i])>0) 
          {
            pattot_in_frame[obsevents.frame[i]] +=
                             1.0/abs(float(obsevents.pattern[i])); 
            if (obsevents.pattern[i] >= 1 && obsevents.pattern[i] <=4) {
               good_pattot_in_frame[obsevents.frame[i]] +=
                             1.0/abs(float(obsevents.pattern[i]));
            }
          }
     }
     // Set the number of events and good events in each frame as an integer
     for (int i=0; i<=obsevents.frame[N-1]; ++i) {
         nevents_in_frame[i] = floor(pattot_in_frame[i]+0.5);
         good_nevents_in_frame[i] = floor(good_pattot_in_frame[i]+0.5);
     }

     errstr<<"Events in frame [60,63] "<<pattot_in_frame[60]<<" "
           <<nevents_in_frame[60]<<" "<<pattot_in_frame[63]<<" "
           <<nevents_in_frame[63]<<" "<<good_nevents_in_frame[63]
           << Sas::message(Msg::App, Msg::Verbose);

     // Find the maximum number of events in a single frame
     int maxevents = 0;
     for (unsigned n=0; n<nevents_in_frame.size(); ++n)
         maxevents = max(nevents_in_frame[n], maxevents);

     errstr<<"Max number of events in one frame: "<<maxevents
           << Sas::message(Msg::App, Msg::Verbose);

     // Define a vector to hold the number of frames holding n events
     vector<int> nframes_with_nevents(maxevents+1,0); // Zero it !

     // Count the number of frames containing n events
     for (unsigned f=0; f<nevents_in_frame.size(); ++f)
         nframes_with_nevents[nevents_in_frame[f]]++;         

     errstr<<"nframes_with_nevents: "<<nframes_with_nevents[0]<<" "
           <<nframes_with_nevents[1]<<" "<<nframes_with_nevents[2]<<" "
           <<nframes_with_nevents[3]<<" "<<nframes_with_nevents[4]
           <<" TOTAL: "<<nevents_in_frame.size()
           << Sas::message(Msg::App, Msg::Verbose);
     
     // Set the number of 1 event frames
     _N0EventFrames = nframes_with_nevents[0];
     _N1EventFrames = nframes_with_nevents[1];

     // Define a 2-d vector to hold the indices of charges in frames
     // which hold n events
     vector<vector<int> > ind_of_charges_with_nevents_in_frame(maxevents+1); 

     // Loop over each frame and assign the indices to a vector element
     for (unsigned i=0; i<obsevents.frame.size(); ++i)
         ind_of_charges_with_nevents_in_frame[nevents_in_frame[
                                obsevents.frame[i] ]].push_back(i);

     // Define a temporary array to hold indices of charges which are 
     // wanted after frame normalisation - and a new frame counter
     vector<unsigned> tempindex, frame_out;

     int out_frame_counter=0;

     // Loop over the number of events in a frame starting with 2.
     for (int nev=2; nev<=maxevents; ++nev) {

        /* We now save n frames which have (nev-1) events, where n 
           is the number of frames which have nev events. If there are 
           less frames with (nev-1) events than there are with nev events
           we duplicate to make the numbers up */
         if ( nframes_with_nevents[nev-1]>1 && nframes_with_nevents[nev]>1) 
          { // i.e. there are frames with this number of events to consider

            int f=0; // frame counter for this value of nev

            // Loop until the number of required frames is reached 
            while (f<nframes_with_nevents[nev])
            {
                int current_frame=-1;

              // Loop over all charges in frames with nev-1 events
                for (unsigned i=0; 
                     i<ind_of_charges_with_nevents_in_frame[nev-1].size(); ++i)                 {
                    // Add the index of this charge to a temp array
                    int ti = ind_of_charges_with_nevents_in_frame[nev-1][i];
                    tempindex.push_back(ti);

                    // Increment frame counters if this is a new frame
                    if (obsevents.frame[ti] != current_frame) {
                       current_frame=obsevents.frame[ti];
                       ++f;
                       ++out_frame_counter;

                       // Sum up the weights of the frames which will be used
                       _sum_of_frame_weights += 
                             good_nevents_in_frame[current_frame];
                    }

                    // set the output frame counter for this charge
                    frame_out.push_back(out_frame_counter);

                    // Exit the loop if the required number of frames reached
                    if (f>=nframes_with_nevents[nev]) {
                       //cout<<"BRK: "<<nev<<" "<<f<<" "<<nframes_with_nevents[nev]<<endl;
                       break;
                    }

                } // End of loop over charges with this number of events in frm
            }
          } // End of if events for this number of events block

     } // end of loop over number of events in frame

     errstr<<"Normalised frame number: "<<frame_out[0]<<" "<<frame_out[1]
           << Sas::message(Msg::App, Msg::Noisy);

     // Add on the one-event frames to the sum of the frame weights
     _sum_of_frame_weights += _N1EventFrames; // these have a weighting of 1.

     // Take the event indices calculated before and write the events into
     // an output structure
     for (unsigned i=0; i<tempindex.size(); ++i) {
         frame_normalised_events.x.push_back(obsevents.x[tempindex[i]]);
         frame_normalised_events.y.push_back(obsevents.y[tempindex[i]]);
         frame_normalised_events.charge.push_back(obsevents.charge[tempindex[i]]);
         frame_normalised_events.pattern.push_back(obsevents.pattern[tempindex[i]]);
         frame_normalised_events.event_id.push_back(frame_out[i]);
         frame_normalised_events.frame.push_back(frame_out[i]);
         frame_normalised_events.weight.push_back(
                            good_nevents_in_frame[obsevents.frame[tempindex[i]]]);
     }

     //cout<<"Norm events: "<<frame_normalised_events.pattern.size()<<endl;
     return frame_normalised_events;

    }

/*
* Method: readEvents(string & eventset)
* Purpose: Reads charges from the raw event file for a CCD
*/
  PiledUpResponseCorrector::ObsEvents PiledUpResponseCorrector::readEvents(const string & eventfile)
   {

   /* First of all we need to process the raw event file to add the
      time information to each event and convert PHA to PI.
    *
    * For now we leave this to the user. At a later stage we will
    * perform the raw processing within this class 
   */
    // processRawData(eventfile);

   // Create empty events structure 
     ObsEvents obsevents;

   // Open the file - NB this is an intermediate file produced by epevents
     DataSet *eventset = dataSetServer->open( eventfile, DataSet::Read);
     
   // Check eventfile contains events extension
     Table *table=0;
     if (eventset->hasBlock("EVENTS")) // SORT OUT names with OAL
      {
        table = eventset->table("EVENTS");
      } 
     else {
        errstr << "Extension EVENTS not found in raw event file "
                 << eventfile<<" "<<Sas::error("invalidEventfile");
      }

   // Read event columns 
     vector<double> event_times;
     vector<int> frame, x, y, charge, pat_typ;
     CalOalUtility::fillVector(table, "TIME", event_times);
     CalOalUtility::fillVector(table, "FRAME", frame);
     CalOalUtility::fillVector(table, "RAWX", x);
     CalOalUtility::fillVector(table, "RAWY", y);
     CalOalUtility::fillVector(table, "PHA_CTI", charge);
     CalOalUtility::fillVector(table, "PAT_TYP", pat_typ);

   // NB: This event table MUST be ordered on time - check here
     checkTimeOrdering(event_times, eventfile);
   
   // Each charge is recorded in the PHA_CTI column but in units of ADU.
   // This needs to be converted into PI by multiplying by 5.
     for (unsigned n=0; n<charge.size(); ++n)
         charge[n] = charge[n] * 5;  

   // Calculate ranges of indices of events which lie within the times used to 
   // extract the spectrum
     list<pair<int,int> > index_range = findEventRanges(event_times);

   // Find the total number of frames which have been exposed
     _totalFrames = getNumberofExposedFrames(eventset);

   // Find frame ranges from the indices
    // _framesToFill = getFramesToFill(frame, index_range, totalFrames);

   // Find the number of frames
     int ilow=0, ihigh=0;
     //RangeBound lowerB_index, upperB_index;

     for ( list<pair<int,int> >::iterator i = index_range.begin();
                                        i != index_range.end(); ++i )
         {
            //(*i).values(ilow, ihigh, lowerB_index, upperB_index);
            ilow = (*i).first;
            ihigh = (*i).second;

            errstr<<"Index range: "<<ilow<<" "<<ihigh
                  << Sas::message(Msg::App, Msg::Verbose);

           // Write each of the required charges into the output structure
            for (int n=ilow; n<=ihigh; ++n) 
            {
               obsevents.event_id.push_back(-1);
               obsevents.pattern.push_back(pat_typ[n]);
               obsevents.charge.push_back(charge[n]);  
               obsevents.x.push_back(x[n]);            
               obsevents.y.push_back(y[n]);            
               obsevents.frame.push_back(frame[n]);            
            }
         }

   // Close event file
     dataSetServer->close(eventset);

     return obsevents;
   }

/*
* Method: extractCharges()
* Purpose: Remove charges which lie outside the extraction region
*/
  PiledUpResponseCorrector::ObsEvents 
       PiledUpResponseCorrector::extractCharges(PiledUpResponseCorrector::ObsEvents obsevents)
   {
     // Make a PSF which just includes the extraction region
      PattFracs *patternFractions = new PattFracs(_cenRaw); 

     // Get the plate scale for this instrument
      MiscDataServer * mds = 0; mds = calServer.getAtom(mds);
      double plateX = mds->get("PLATE_SCALE_X");
      double plateY = mds->get("PLATE_SCALE_Y");
      calServer.releaseAtom(mds);
   
      PsfImage *psfim = new PsfImage(1000.0, patternFractions,
                               _cenTel.theta(), _cenTel.phi());
      psfim->cleanRegion(_pxitr, _cenDet);
      Image<float> &pim = psfim->image(0);
      int nx = pim.xExtent(), ny = pim.yExtent();

     // Find range of RAWX, RAWY values in PSF
      int xmin = floor ( ( 0 - floor(nx/2) ) / plateX + 0.5) + _cenRaw.x();
      int xmax = floor ( ( (nx-1) - floor(nx/2) ) / plateX + 0.5) + _cenRaw.x();      
      int ymin = floor ( ( 0 - floor(ny/2) ) / plateY + 0.5) + _cenRaw.y();
      int ymax = floor ( ( (ny-1) - floor(ny/2) ) / plateY + 0.5) + _cenRaw.y();      
      Image<int> RawIm((xmax-xmin+1),(ymax-ymin+1));
      for (int x=0; x<(xmax-xmin+1); ++x)  
       for (int y=0; y<(ymax-ymin+1); ++y)
           RawIm[x][y]=0;

     // Convert from PSF pixels to a yes/no RAW pixels image
      int ix=0, iy=0;
      for (int x=0; x<nx; ++x) { // Loop over PSF pixels
       for (int y=0; y<ny; ++y) {
         // Convert to RAW pix
          ix = floor ( ( x - floor(nx/2) ) / plateX + 0.5) + _cenRaw.x();
          iy = floor ( ( y - floor(ny/2) ) / plateY + 0.5) + _cenRaw.y();
          if (pim[x][y]>0.0)
              RawIm[ix-xmin][iy-ymin]=1;
       }
      }

     // Make a vector of accepted events
      ObsEvents obsevents_out;
      for (unsigned n=0; n<obsevents.x.size(); ++n) {

          if (obsevents.x[n] >= xmin && obsevents.x[n] <= xmax &&
              obsevents.y[n] >= ymin && obsevents.y[n] <=ymax) {

            if (RawIm[obsevents.x[n]-xmin][obsevents.y[n]-ymin]==1) {
               obsevents_out.event_id.push_back(obsevents.event_id[n]);
               obsevents_out.pattern.push_back(obsevents.pattern[n]);
               obsevents_out.charge.push_back(obsevents.charge[n]);
               obsevents_out.x.push_back(obsevents.x[n]);
               obsevents_out.y.push_back(obsevents.y[n]);
               obsevents_out.frame.push_back(obsevents.frame[n]);
            }
         }
      }

      delete psfim; psfim=0;

      return obsevents_out;
   }

/*
* Method: expandCharges()
* Purpose: Find the ranges of events corresponding to the ranges of times
*          used to extract the spectrum
*/
const vector<int> PiledUpResponseCorrector::expandCharges(vector<int> pha, 
                                                  vector<int> chargeIn,
                                                  vector<int> pat_type) const
   {
      vector<int> charge;

     // Loop over each entry in the raw event file
      unsigned i=0;
      while (i < pha.size())
      {

       // Find how many pixels (charges) made this event
        unsigned pt = abs(pat_type[i]);

       // If this charge has a type of 0,1 or -1 it is a single-pixel event 
        if (pt <= 1) {

          // Write charge directly into the output array
           charge.push_back(chargeIn[i]);
           i++;

        } else { // Multi-pixel event

           int phatot=0, pitot=0;

          // Loop over charges belonging to this event
           for (unsigned n=0; n<pt; ++n)
            {
             // Sum the PHAs for this event
              phatot += pha[i+n];

             // Sum the PIs for this event if not NULL
              if (chargeIn[i+n]>0) pitot += chargeIn[i+n];
            }

          // Calculate the PI for each charge in this event
           for (unsigned n=0; n<pt; ++n)
              charge.push_back(pitot * pha[i+n] / phatot);

          // Increment the counter by the number of charges added
           i += pt;
        }
      }
      //cout<<"Expended charges: "<<charge[0]<<" "<<charge[11]<<" "<<charge[15]<<" "<<charge[16]<<" "<<charge[8130]<<" "<<charge[8131]<<endl;

      return charge;
   }

/*
* Method: findEventRanges(vector<double> event_times> >)
* Purpose: Find the ranges of events corresponding to the ranges of times
*          used to extract the spectrum
*/
  list<pair<int,int> > PiledUpResponseCorrector::findEventRanges(
                   vector<double> event_times) 
   {

     list<pair<int,int> > index_range;

    /* Loop over each GTI and find the frame just after the begining and 
     * just before the end to get a contiguous set of frames to be included
     * This isn't perfect because it will miss any frames at the extremes
     * of the GTI which don't have an event but this should be a small number.
     * If no time info then frame range is minimum and maximum
    */
     if (_spectrum_times.begin() == _spectrum_times.end()) 
     {
        index_range.push_back(pair<int,int>(0, event_times.size()-1));
        
     } else {

    // Iterate over each time range
        double tlow=0.0, thigh=0.0;
        //RangeBound lowerB_time, upperB_time;
        int a=0, i1=0, i2=0;
        bool lowfound=false;

        //for ( list<Range<double> >::iterator i = _spectrum_times.begin();
          //                      i != _spectrum_times.end(); ++i )
        for ( unsigned i=0; i<_spectrum_times.size(); ++i )
         {
            tlow = _spectrum_times[i].start();
            thigh = _spectrum_times[i].stop();

            errstr<<"_spectrum_times: "<<tlow<<" "<<thigh
                  << Sas::message(Msg::App, Msg::Verbose);

           // Find the index of the event which has a time = or just after
           // the low time and = or just before the end time
            while ( (a < INT(event_times.size())) &&
                        (event_times[a]<=thigh) ) {

          // Check if this is just greater than the low time
               if ( (event_times[a]>=tlow) && (not lowfound)) {
                  i1=a;
                  lowfound=true;
               }
               ++a;
            }

          // Write out this index range if successfully found
            if (lowfound) {
               i2 = a-1;
               index_range.push_back(pair<int,int>(i1, i2));
               lowfound=false;
            }
         } // End of loop over spectrum time ranges
     } // End of main block

     return index_range;
   }

/*
* Method: getFramesToFill(vector<int> frame, vector<Range<int> > index_range))
* Purpose: Find the frames corresponding to input events
*/
  vector<int> PiledUpResponseCorrector::getFramesToFill(
                                 vector<int> frame, 
                            list<pair<int,int> > index_range, int totalFrames) 
   {

     vector<int> FramesToFill;

    // Find all the frames which have events
     int ilow=0, ihigh=0;
     for ( list<pair<int,int> >::iterator i = index_range.begin();
                                        i != index_range.end(); ++i )
         {
            ilow=(*i).first;
            ihigh=(*i).second;

           // Add frames between these limits to the frames to fill list
            int saveframe=-1;
            for ( int f=ilow; f<=ihigh; ++f)
                if (frame[f] != saveframe) {
                   FramesToFill.push_back(frame[f]);
                   saveframe = frame[f];
                }
         }

     int nf = FramesToFill.size();

     errstr<<"Test frame range: "<<nf<<" frames: "<<frame[ilow]
           <<" "<<frame[ihigh]
           <<"Total frames to fill: "<<FramesToFill.size()<<" "<<totalFrames
           <<"Last frame to fill: "<<FramesToFill[FramesToFill.size()-1]
           << Sas::message(Msg::App, Msg::Verbose);

     return FramesToFill;
   }

/*
* Method: getFramesToFill(vector<int> frame, vector<Range<int> > index_range))
* Purpose: Find the frames corresponding to input events
*/
  vector<int> PiledUpResponseCorrector::getFramesToFill(ObsEvents obsevents,
                                           int N1EventFrames)
{
      vector<int> FramesToFill;

     // Put the frames in order
      vector<int> order = getOrderedIndex(obsevents.frame);
 
      int saveframe=0;
      int size = order.size();

      for ( int i=0; i<size; ++i) {
             //cout<<"O: "<<i<<" "<<order[i]<<" "<<obsevents.frame[order[i]]<<endl;
          if (obsevents.frame[order[i]] > saveframe) {
             //cout<<"O:: "<<i<<" "<<order[i]<<" "<<obsevents.frame[order[i]]<<endl;
             FramesToFill.push_back(obsevents.frame[order[i]]);
             saveframe=obsevents.frame[order[i]];
          }
      }
      //cout<<"ORDER: "<<order.size()<<" "<<FramesToFill.size()<<endl;

      //cout<<"Adding on "<<(totalFrames-saveframe)<<" frames "<<FramesToFill.size()<<endl;
      // Add on extra frames to simulate the frames with one event only
      for ( int i=saveframe+1; i<=saveframe+N1EventFrames; ++i) 
          FramesToFill.push_back(i);

      //cout<<"Total frames to fill: "<<FramesToFill.size()<<" "<<N1EventFrames<<" "<<saveframe<<endl;
      //cout<<"First four: "<<FramesToFill[0]<<" "<<FramesToFill[1]<<" "<<FramesToFill[2]<<" "<<FramesToFill[3]<<" "<<FramesToFill[4]<<" "<<endl;

      return FramesToFill;
}

/*
* Method: getNumberofExposedFrames(eventset)
* Purpose: Find the total number of frames which have been exposed
*/
int PiledUpResponseCorrector::getNumberofExposedFrames(DataSet* eventset) 
{
   // Open the exposure block from the raw event file
     Table *table=0;
     if (eventset->hasBlock("EXPOSURE")) // SORT OUT names with OAL
      {
        table = eventset->table("EXPOSURE");
      }
     else {
        errstr << "Extension EXPOSURE not found in raw event file "
               <<Sas::error("invalidEventfile");
      }

   // Read the times array
     vector<double> exp_times;
     CalOalUtility::fillVector(table, "TIME", exp_times);

   // Subset with the times used to make the spectrum
     list<pair<int,int> > index_range = findEventRanges(exp_times);

   // Find the total number of frames which have been exposed
     int nexposures=0;
     for ( list<pair<int,int> >::iterator i = index_range.begin();
                                        i != index_range.end(); ++i )
         {
            nexposures += (*i).second - (*i).first + 1;
         }
 
     errstr<<"Total number of exposures in raw event file: "<<nexposures
           << Sas::message(Msg::App, Msg::Verbose);

     return nexposures;
}

/*
* Method: mergeEvents(ObsEvents events1, ObsEvents events2)
* Purpose: To Merge together and sort two event lists
*/
  const PiledUpResponseCorrector::ObsEvents 
           PiledUpResponseCorrector::mergeEvents(ObsEvents events1, 
                                                 ObsEvents *testevents) const
   {

     ObsEvents mergedevents;

    // We are only interested in frames where there are events in both
    // lists however we need to keep the others to get the full RMF
    // out at the end. Possible time saving here....

    // The algorithm only works if both event arrays are sorted by frame
    // testevents is sorted by frame by definition. Here we find the 
    // index of each obs event in frame order
     vector<int> order = getOrderedIndex(events1.frame);

     int minx=0, miny=0, maxx=0, maxy=0, charge=0, pattern=0;

    // Loop over each frame
     int a=0;

    // DBG - counters for each type of output pixel
     int nss=0, nsd=0, ndd=0, nsb=0, ndb=0;

     //cout<<"TEST events frames: "<<testevents->frame[0]<<" "<<testevents->frame[1]<<" "<<testevents->frame[2]<<" "<<testevents->frame[testevents->x.size()-1]<<" "<<testevents->x.size()<<endl;
     //cout<<"OBS events frames: "<<events1.frame[0]<<" "<<events1.frame[1]<<" "<<events1.frame[2]<<" "<<events1.frame[events1.x.size()-1]<<" "<<events1.x.size()<<endl;
     //cout<<"OBS events frames: "<<events1.frame[order[0]]<<" "<<events1.frame[order[1]]<<" "<<events1.frame[order[events1.x.size()-1]]<<" "<<events1.x.size()<<endl;

     for (unsigned n=0; n<testevents->x.size(); ++n) {

             //int DBG_check=0;  // REMOVE
        // Set values for this event
         charge = testevents->charge[n];
         pattern = testevents->pattern[n];

        // Assign a position for the secondary pixel if a DOUBLE pixel
        // event
         if (pattern==CombinedEventGrades::DOUBLE) {

            minx = testevents->x[n];
            miny = testevents->y[n];
            maxx = testevents->x[n];
            maxy = testevents->y[n];

            float randir = rand();
            if (randir < -0.25)
               minx=testevents->x[n]-1;
            else if (randir < 0.0)
               maxx=testevents->x[n]+1;
            else if (randir < 0.25)
               miny=testevents->y[n]-1;
            else
               maxy=testevents->y[n]+1;
         }

     // Loop over the frame-ordered index of each added charge 
         while (a < INT(events1.frame.size()) && 
                        events1.frame[order[a]]<=testevents->frame[n]) {

           int oa = order[a]; // The position of the a'th frame 

           //cout<<"Trial: "<<events1.frame[order[a]]<<" "<<events1.x[order[a]]<<" "<<events1.y[order[a]]<<endl;
          // Check if this charge is from the right frame
           if (events1.frame[oa]==testevents->frame[n]) {

            // Set the weight for this test event from the frame weight
             testevents->weight[n] = events1.weight[oa];

            // DEBUG t oremove
             //if (events1.weight[oa]<1) {
             //DBG_check=1;
             //cout<<"orig weight: "<<events1.weight[oa]<<" "<<events1.frame[oa];
             //}
            // DEBUG t oremove

            // SINGLE
             if (pattern==CombinedEventGrades::SINGLE) {

              int xdist = events1.x[oa] - testevents->x[n];
              int ydist = events1.y[oa] - testevents->y[n];
              int totdist = (abs(xdist) + abs(ydist));

             // Adding charge onto single pixel
              if ( totdist == 0 ) {

                 charge += events1.charge[oa];

                 nss++;

             //    cout<<"SinToSin: "<<events1.frame[oa]<<" "<<testevents->x[n]<<endl;

             // Creating a double from a single
              } else if ( totdist == 1 ) {

                 pattern = CombinedEventGrades::DOUBLE;
                 charge += events1.charge[oa];
                 
                 minx = min(testevents->x[n],events1.x[oa]);
                 maxx = max(testevents->x[n],events1.x[oa]);
                 miny = min(testevents->y[n],events1.y[oa]);
                 maxy = max(testevents->y[n],events1.y[oa]);

                 nsd++;
               //  cout<<"SinToDub: "<<events1.frame[oa]<<" "<<testevents->x[n]
                 //    <<" "<<testevents->y[n]<<" "<<events1.x[oa]<<" "<<
                   //events1.y[oa]<<" "<<charge<<" "<<events1.charge[oa]<<" "<<
                // minx<<" "<<maxx<<" "<<miny<<" "<<maxy<<endl;
             // Diagonal touching pixel
              } else if ( totdist == 2) {

                 pattern = 999;
                 nsb++;

               //  cout<<"SinToBad: "<<events1.frame[oa]<<" "<<testevents->x[n]<<endl;
                 break;

              }

           // DOUBLE - either originally or after a single becomes double
             } else { 

            // Test if this added event pixel falls on top of a test charge
              if ( (events1.x[oa] == minx && events1.y[oa] == miny ) ||
                   (events1.x[oa] == maxx && events1.y[oa] == maxy ) ) {
                
                 charge += events1.charge[oa]; 

                 ndd++;

             // Or if it produces a bad pixel
              } else if ( (events1.x[oa] >= (minx-1)) &&
                       (events1.x[oa] <= (maxx+1)) &&
                       (events1.y[oa] >= (miny-1)) &&
                       (events1.y[oa] <= (maxy+1)) ) {
                
                 pattern = 99;  // Set pattern bad

               //  cout<<"DubToBad: "<<events1.frame[oa]<<" "<<testevents->x[n]<<endl;
                 ndb++;
                 break;
              }

           } // End of SINGLE or DOUBLE pixel test event block

          } // End of whether this event has the right frame 
        
          // Increment the added event counter
          ++a;

         } // End of while loop over added charges with this frame number

       // Remove one from the weight if the test event has piled-up
         if (charge > testevents->charge[n]) --testevents->weight[n];

         //if (DBG_check > 0) 
           // cout<<"testevents->weight[n]: "<<pattern<<" "<<testevents->weight[n]<<endl;

       //  cout<<"Nevents: "<<testevents->frame[n]<<" "<<a<<" "<<events1.x[a]<<" "<<events1.y[a]<<" "<<events1.frame[a]<<endl;

        // Write output event (energy, pattern, testevents.x, testevents.y)
         if (pattern == CombinedEventGrades::SINGLE ||
             (pattern == CombinedEventGrades::DOUBLE &&
              _eventGrade != CombinedEventGrades::SINGLE) ) {

            mergedevents.frame.push_back(testevents->frame[n]);
            mergedevents.x.push_back(testevents->x[n]);
            mergedevents.y.push_back(testevents->y[n]);
            mergedevents.charge.push_back(charge);
            mergedevents.pattern.push_back(pattern);
            mergedevents.event_id.push_back(testevents->event_id[n]);
            mergedevents.weight.push_back(testevents->weight[n]);

         //   cout<<"EVT: "<<testevents->frame[n]<<" "<<testevents->x[n]<<" "<<testevents->y[n]<<endl;

         }
    } // End of loop over test events

   //*DEBUG
    //for (int i=0; i<=10; ++i)
      //  cout<<mergedevents.frame[i]<<" "<<mergedevents.x[i]
        //    <<" "<<mergedevents.y[i]<<" "<<mergedevents.charge[i]
          //  <<" "<<mergedevents.pattern[i]<<endl;

    //cout<<"NSS: "<<testevents->x.size()<<" "<<nss<<" "<<nsd<<" "<<nsb<<" "<<ndd<<" "<<ndb<<endl;

   // END DEBUG

    return mergedevents;
  }

/*
* Method: makeRmfRow(ObsEvents events)
* Purpose: Produces an RMF row from an array of events
*/
  const vector<double> PiledUpResponseCorrector::makeRmfRow(
                           PiledUpResponseCorrector::ObsEvents events,
                           unsigned nframes) const
   {

    // Initialise the output vector
     vector<double> correctedResponse(_response->numberOfChannels(),0);

    // Define the bin factor - CHECK that the last, first and number of chans ok

     errstr<<"BINfactor: "<<_chanWidth<<" "<<_chanOffset
           << Sas::message(Msg::App, Msg::Verbose);

//     cout<<"Weights: "<<events.weight[1000]<<" "<<_sum_of_frame_weights<<endl;

    // Make a weighted histogram of the PI channels of the input events
     for (unsigned n=0; n<events.charge.size(); ++n) {

      // Subset here if 
        // Assign event to a channel bin
         unsigned i = INT( (events.charge[n] - _chanOffset ) /
                                      _chanWidth);

         if (i < correctedResponse.size())
            //correctedResponse[i]++;
            correctedResponse[i] += float(events.weight[n]);
     }

     //cout<<"Weights2: "<<correctedResponse[100]<<" "<<_sum_of_frame_weights<<endl;

    // Normalise the histogram by the number of test events weighted by the
    // number of good events in each frame
     double dbgtot=0.0;

     for (unsigned i=0; i<correctedResponse.size(); ++i) {
         //correctedResponse[i] /= float(nframes);
         correctedResponse[i] /= float(_sum_of_frame_weights);
         dbgtot += correctedResponse[i];
     }
     
    // Output the normalisation for this energy
     errstr<<"RMF normalisation = "<<dbgtot
           << Sas::message(Msg::App, Msg::Verbose);
     
    // Return the response normalised to the number of original test events
     return correctedResponse;
   }

/*
* Method: checkTimeOrdering(event_times)
* Purpose: Makes sure that the raw events are ordered in ascending time
*/
  void PiledUpResponseCorrector::checkTimeOrdering(
                                      vector<double> event_times, 
                                      const string &eventfile) const
   {
     for ( unsigned i = 1; i<event_times.size(); ++i) {
         if (event_times[i] < event_times[i-1]) {
            errstr << "Raw event file "<<eventfile<< " not ordered in time"
                   <<Sas::error("invalidEventfile");
         }
     }
   }

/*
* Method: getOrderedIndex()
* Purpose: Orders event into frame order
*/
  const vector<int> PiledUpResponseCorrector::getOrderedIndex(
                                   vector<int> v) const
 {

  // Use the trick of converting the input vector into a pair
   //vector<pair<int,int> > vpair;
   vector<pairInt > vpair;
   int size=v.size();

   for ( int i = 0; i<size; ++i)  
       vpair.push_back(pairInt(v[i],i));
  
   std::sort(vpair.begin(), vpair.end());

   vector<int> order;
   for ( int i = 0; i<size; ++i)
      order.push_back(vpair[i].i2);

   return order;
 }

/*
* Method: processRawData()
* Purpose: Processes the raw event file into a file that
*          contains the FRAME, TIME, RAWX, RAWY and PI values 
*          of each charge comprising an event
*/
/*
string PiledUpResponseCorrector::processRawData(const OdfSet* os)
{

   // Should find the raw event file to use automatically from the CCD
   // number.
    string rawEventFile = ...
    string gtifile = 'tempgti.dat'
    string tempeventset = 'tempevents.dat'
    string eventset = 'piledevents.dat'

   // Run epframes to associate frame number with time
    Info::Task* epframes = task("epframes");
    ParameterInfo& epframespars = *epframes->parameters();
    epframespars.revertToDefault();
    epframespars["set"]->setValue(rawEventFile);
    epframespars["gtiset"]->setValue(gtifile);
    epframespars["eventset"]->setValue(tempeventset);

    if(!ePS_->proc.protectedrun(epframes))
      throw(ProcException(""));

   // Run epevents to convert PHA into PI and maintain the charges
    Info::Task* epevents = task("epevents");
    ParameterInfo& epeventspars = *epevents->parameters();
    epeventspars.revertToDefault();
    epeventspars["eventset"]->setValue(tempeventset);
    epeventspars["outset"]->setValue(eventset);

    if(!ePS_->proc.protectedrun(epevents))
      throw(ProcException(""));

    return eventset;  
}
*/
