// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// This  skeleton file was automatically created by 
//            pkgmaker 0.1
//            Thursday, December 11 1997
// 
// Module: PiledUpResponseCorrector
// Author: Hassan Siddiqui
//
// Description: 
// Corrects response for pile-up
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: PiledUpResponseCorrector.h,v 1.3 2013/04/04 13:16:02 rsaxton Exp $

#ifndef PiledUpResponseCorrector_H
#define PiledUpResponseCorrector_H

// General classes

#include <iostream>
#include <vector>
#include <map>

// SAS-specific classes

#include "Dal.h"
#include "Tracer.h"
#include "Xmm.h"
#include "PixCoords.h"

// Local Classes
#include "GenericResponse.h"
#include "Frame.h"
#include "ResponseCorrector.h"
#include "CorePosData.h"
#include "PattFracs.h"
#include "RmfUtils.h"


// -----------------------------------------------------------------------------------

// class PiledUpResponseCorrector

// Corrects response row for pile-up

// -----------------------------------------------------------------------------------

class PiledUpResponseCorrector : public ResponseCorrector
{

public:
  
  /* Constructor */

  PiledUpResponseCorrector( const GenericResponse  * response,
                            const string & eventfile,
                            CorePosData   * corePosData,
                            unsigned      chanoffset,  unsigned chanwidth,
                            PixelIterator * pxitr);

  ~PiledUpResponseCorrector();

  /* Methods */

/*
  const void initialise(string & eventfile,
                         PixCoords cenRaw, TelCoords cenTel,
                         CoordPair cenDet, EventGrades eventGrade,
                         unsigned      chanoffset,  unsigned chanwidth,
                         PixelIterator * pxitr);
*/
  const vector<double> correctedResponse( const Interval<double>& ebin0,
                                  const vector<double>& response  ) const;

  // Events structure
  typedef struct
    {
      vector<int> frame;      // The frame number for each charge
      vector<int> x;          // The raw X coord for this charge
      vector<int> y;          // The raw Y coord for this charge
      vector<int> charge;     // Charge in this pixel (ADU)
      vector<int> pattern;    // The pattern of the event of this charge
      vector<int> event_id;   // An id used internally
      vector<int> weight;     // A weight assigned to this event
    } ObsEvents;


private:

  /* Internal methods */
  ObsEvents * testEvents(double energy0, const vector<double> response0) const;

  ObsEvents normalise_frames(ObsEvents obsevents);

  const vector<CoordPair> getChipPosition(
                                int nevents, double energy0, 
                                PattFracs *patternFractions) const;

  const vector<int> findrand(int N, vector<double> probs,
                             vector<int> values) const;

  const vector<int> findrand(int N, vector<double> probs) const;

  template<class T>
  const vector<int> findrand2(int N, T *probs, unsigned nprobs) const;

  ObsEvents readEvents(const string & eventfile);

  ObsEvents extractCharges(ObsEvents obsevents);

  const ObsEvents mergeEvents(ObsEvents events1, ObsEvents *testevents) const;

  const vector<double> makeRmfRow(ObsEvents events, unsigned nframes) const;

  list<pair<int,int> > findEventRanges(vector<double> event_times);

  vector<int> getFramesToFill(vector<int> frame, 
                           list<pair<int,int> > index_range, int totalFrames);
  //vector<int> getFramesToFill(ObsEvents obsevents, int NonechargeFrames);
  vector<int> getFramesToFill(ObsEvents obsevents, int totalFrames);

  int getNumberofExposedFrames(DataSet* eventset);

  void checkTimeOrdering(vector<double> event_times, 
                                       const string & eventfile) const;

  const vector<int> getOrderedIndex(vector<int> v) const;

  const vector<int> expandCharges(vector<int> pha, vector<int> chargeIn,
                                             vector<int> pat_type) const;

  /* Data members */
  const GenericResponse    *  _response;
  ObsEvents _obsevents;   // Charges from input event file after normalisation
  PixCoords _cenRaw; // Centre of PSF in raw coords
  TelCoords _cenTel; // Centre of extraction region in theta, phi       
  CoordPair _cenDet; // Centre of extraction region in detector coords
  EventGrades _eventGrade;// Patterns for this spectrum
  GtiList _spectrum_times; // Times used to extract the spectrum 
  unsigned _chanOffset,   // PI channel of left hand side of 0th PI bin
           _chanWidth;    // Width of each PI bin
  PixelIterator * _pxitr; // An iterator for pixels within the extraction region
  int      _nframes;      // Total number of frames from event file
  vector<int> _framesToFill; // Numbers of exposed frames to be considered
  //int _NonechargeFrames;  // Number of frames with one event
  int _totalFrames;  // Total number of frames in the event file
  int _N0EventFrames;  // Number of frames with no events
  int _N1EventFrames;  // Number of frames with one event
  int _sum_of_frame_weights; // Sum of the weights of all the input frames

  static const unsigned _pn_adu_to_pi = 5; // COnversion from ADU to PI channel

  /* Functions */

  /* Disable default constructor */
  PiledUpResponseCorrector();

};

class pairInt 
{
public:
  pairInt(int i1, int i2) {
      this->i1=i1; this->i2=i2;
  }
  int i1, i2;
};

inline bool operator <( const pairInt& l, const pairInt& r)
     { return l.i1 < r.i1; } 


#endif






